module Main where
data V = VUnit | VPair V V | VL V | VR V

instance Show V where
    show VUnit = "()"
    show (VPair x y) = "(" ++ show x ++ ", " ++ show y ++ ")"
    show (VL v) = "Inl " ++ show v
    show (VR v) = "Inr " ++ show v

{-
vs :: V -> V
vs v = VL v

v0 :: V
v0 = VR VUnit

v1 = vs v0

v2 = vs v1

plus :: V -> V -> V
plus x (VL z) = VL (plus x z)
plus x _ = x

mult :: V -> V -> V
mult x (VL z) = plus x (mult x z)
mult x y = y

mult2 (VL z) y = plus (mult z y) y
mult2 x _ = x

rev x = rh x (VR VUnit)

rh (VL (VPair e r)) s = rh r (VL (VPair e s))
rh _ s = s

vnil = VR VUnit
vcons x y = VL (VPair x y)
-}

type Surface = (Bool, (Int, (Int, Int)))

surf :: Surface
surf = (True, (1, (1, 1)))

mkSurf :: Bool -> Int -> Int -> Int -> Surface
mkSurf d r t e = (d, (r, (t, e)))

main :: IO ()
main =
    do --putStr $ unlines $ map show $ initr [surf, surf]
      let input =
              [mkSurf True 0 0 0,
               mkSurf True 0 0 2]
      putStrLn $ show $ rmain input

consList :: a -> [[a]] -> [[a]]
consList x [] = []
consList x (y:ys) = ((x:y):consList x ys)

permute :: [b] -> [a] -> [[a]]
permute [] _ = [[]]
permute (x:xs) ys =
    permute' (permute xs ys) ys

permute' :: [[a]] -> [a] -> [[a]]
permute' xs [] = []
permute' xs (y:ys) = (consList y xs) `app` (permute' xs ys)

app :: [a] -> [a] -> [a]
app [] ys = ys
app (x:xs) ys = x:(app xs ys)

{-
appListLeft :: [a] -> [[a]] -> [[a]]
appListLeft _ [] = []
appListLeft x (y:ys) = (app x y):(appListLeft x ys)

appListRight :: [a] -> [[a]] -> [[a]]
appListRight _ [] = []
appListRight x (y:ys) = (app y x):(appListRight x ys)
-}

leftThree :: [(Int, Int)]
leftThree = [(0, 0), (1, 0), (2, 0)]

--rightThree :: [(Int, Int)]
--rightThree = [(0, 0), (0, 1), (0, 2)]

nine :: [(Int, Int)]
nine = [(0, 0), (0, 1), (0, 2),
        (1, 0), (1, 1), (1, 2),
        (2, 0), (2, 1), (2, 2)]

makeGuesses :: [Surface] -> [[(Int, Int)]]
makeGuesses ss =
    permute' (permute ss nine) leftThree

{-
rmain :: [Surface] -> Int
rmain ss =
    let gs = makeGuesses ss
    in checkAll ss gs
-}

checkAll :: [Surface] -> [[(Int, Int)]] -> Int
checkAll ss (g@((l0, _):_):gs) =
    let result = check ss g
    in if result
       then l0
       else checkAll ss gs

check :: [Surface] -> [(Int, Int)] -> Bool
check [] [(_, ln)] =
    ln == 0
check (s1:ss) (r0:r1:rs) =
    error "not implemented"

radd :: Int -> Int -> Int
radd 2 _ = 2
radd _ 2 = 2
radd 1 1 = 2
radd 0 0 = 0
radd _ _ = 1

rmul :: Int -> Int -> Int
rmul 0 _ = 0
rmul _ 0 = 0
rmul 2 2 = 2
rmul _ _ = 1

-- New idea

eq :: Int -> Int -> Bool
eq 0 0 = True
eq 1 1 = True
eq 2 2 = True
eq _ _ = False

raysEq :: [(Int, Int)] -> [(Int, Int)] -> Bool
raysEq [] [] = True
raysEq (x@(xl, xr):xs) (y@(yl,yr):ys) =
    if xl /= yl || xr /= yr
       then False
       else raysEq xs ys

repl :: [a] -> b -> [b]
repl [] _ = []
repl (x:xs) y = y:(repl xs y)

initr :: [Surface] -> [(Int, Int)]
initr ss = (0,0):(repl ss (0,0))

type Result = (Int, [(Int, Int)])

rmain :: [Surface] -> Result
rmain ss =
    let rs = initr ss
    in iter ss rs

iter :: [Surface] -> [(Int, Int)] -> Result
iter ss rs =
    let rs' = step ss rs
    in if raysEq rs rs'
       then finish rs
       else iter ss rs'

finish :: [(Int, Int)] -> Result
finish theList@((l0,_):_) = (l0, theList)

step :: [Surface] -> [(Int, Int)] -> [(Int, Int)]
step [] rs = rs
step (s@(d, (r, (t, e))):ss') (p0@(l0,r0):p1@(l1,r1):rs') =
    let l0' =
            if d
            then (r `rmul` r0) `radd`
                 (t `rmul` l1) `radd`
                 e
            else l1
        r1' =
            if (not d)
            then (r `rmul` l1) `radd`
                 (t `rmul` r0) `radd`
                 e
            else r0
    in (l0',r0):(step ss' ((l1,r1'):rs'))

