module ReadUtil where

import Data.List

readsChar :: ReadS Char
readsChar [] = []
readsChar (c:cs) = [(c, cs)]

readsLine :: ReadS String
readsLine s =
    [("", s') |
     ('\n', s') <- readsChar s]
    ++
    [(c:str, s') |
     (c, s1) <- readsChar s,
     c /= '\n',
     (str, s') <- readsLine s1]

readsWhitespace :: ReadS ()
readsWhitespace (' ':s') = readsWhitespace s'
readsWhitespace ('\t':s') = readsWhitespace s'
readsWhitespace ('\n':s') = readsWhitespace s'
readsWhitespace ('\r':s') = readsWhitespace s'
readsWhitespace s = [((), s)]

removeCommas :: String -> String
removeCommas s = filter (/= ',') s

readsSentences :: ReadS [[String]]
readsSentences s =
    [(map words sentences, s') |
     (sentences, s') <- readsSentences' (removeCommas s)]

readsSentences' :: ReadS [String]
readsSentences' s =
    case findIndex (== '.') s of
      Nothing -> [([], s)]
      Just pos ->
          let left = take pos s
              right = drop (pos+1) s
          in [((left:sentences), s') |
              (sentences, s') <- readsSentences' right]

readsPrompt :: ReadS ()
readsPrompt s =
    [((), s') |
     (_, s1) <- readsWhitespace s,
     ('>', s2) <- readsChar s1,
     (':', s3) <- readsChar s2,
     (' ', s') <- readsChar s3]

