module Room where

import Data.List
import Data.Maybe
import Item
import ReadUtil

type Room = String

data RoomData = RoomData {
    room_name :: Room,
    room_desc :: String
} deriving (Show)

readsRoom :: ReadS (RoomData, [Item])
readsRoom s =
    [((RoomData { room_name = name, room_desc = "" },
       sentencesToItems sentences),
      s') |
     (name, s1) <- readsLine s,
     ('\n', s2) <- readsChar s1,
     (sentences, s3) <- readsSentences s2,
     (_, s') <- readsPrompt s3]

sentencesToItems :: [[String]] -> [Item]
sentencesToItems ss = mapMaybe sentenceToItem ss

sentenceToItem :: [String] -> Maybe Item
sentenceToItem ("There":"is":('a':_):"(broken)":rest) =
    sentenceToItem' (take (length rest - 1) rest)
sentenceToItem ("There":"is":('a':_):rest) =
    sentenceToItem' (take (length rest - 1) rest)
sentenceToItem ("Underneath":"the":rest) =
    sentenceToItem2 rest
sentenceToItem _ = Nothing

sentenceToItem2 :: [String] -> Maybe Item
sentenceToItem2 [] = Nothing
sentenceToItem2 ("there":"is":('a':_):"(broken)":rest) =
    sentenceToItem' rest
sentenceToItem2 ("there":"is":('a':_):rest) =
    sentenceToItem' rest
sentenceToItem2 (_:rest) = sentenceToItem2 rest

sentenceToItem' :: [String] -> Maybe Item
sentenceToItem' s = Just (unwords s)
