module Main where

import Data.Char
import Data.Array
import Data.List
import System.IO
import System.Exit
import Room
import Item
import Dir
import Solve

main :: IO ()
main =
    do hSetBuffering stdin NoBuffering
       hSetBuffering stdout NoBuffering

       login
       start

       say "take pamphlet"
       say "inc pamphlet"

       go North
       solve "keypad"
       go South
       say "use keypad"
       say "inc keypad"

       say "take /etc/passwd"
       say "inc /etc/passwd"
       say "take note"
       say "inc note"

       goAndSolve "power cord" [East, South]

       say "take downloader"

       say "combine downloader power cord"

       goAndSolve "USB cable" [East]
       say "combine downloader USB cable"

       goAndSolve "display" [East, North]
       say "combine downloader display"

       goAndSolve "jumper shunt" [East, South, East]
       say "combine downloader jumper shunt"

       goAndSolve "progress bar" [East, North, North]
       say "combine downloader progress bar"

       
       --say "take downloader"
       --say "inc downloader"

       say "use downloader"
       say "inc downloader"

       go East
       go East
       go East
       play "rs232.txt"
       go West
       go West
       go West

       say "take uploader"

       say "combine uploader RS232 adapter"
       
       goAndSolve "MOSFET" [East, South, East, East]
       say "combine uploader MOSFET"

       goAndSolve "status LED" [East, North, North, East, East]
       say "combine uploader status LED"

       --goAndSolve "RS232 adapter" [East, East, East]
       --say "combine uploader RS232 adapter"

       goAndSolve "EPROM burner" [East, East, East, North]
       say "combine uploader EPROM burner"

       goAndSolve "battery" [East, North, East]
       say "combine uploader battery"

       putStrLn "use uploader"
       code1 <- readFile "code1"
       putStr code1
       putStr "EOM"
       waitPrompt
       return ()

       --sequence_ $ map hackChar [0..111]


       --theList <- sequence $ map hackChar [0..111]
       --hPutStrLn stderr $ unlines $ map ("CHAR:" ++) $ map show theList

       --go North
       --go North

       --play "usb.txt"
       --go West
       --say "take downloader"
       --say "combine downloader usb cable"
       return ()

{-
       (roomData, items) <- look
       hPutStrLn stderr $ show roomData
       hPutStrLn stderr $ unlines $ map show items
       itemDatas <- lookItems items
       hPutStrLn stderr $ unlines $ map show itemDatas

       let trees = makeTrees (Whole "USB cable") itemDatas
       hPutStrLn stderr $ "Num trees = " ++ show (length trees)
       hPutStrLn stderr $ unlines $ map show $ map fst trees
-}

hackChar :: Int -> IO ()
hackChar i =
    do hackUpload i
       hackBit 0 7

hackBit :: Int -> Int -> IO ()
hackBit sofar (-1) =
    do say $ "CHAR:" ++ show (chr sofar)
       return ()
       --return (chr sofar)
hackBit sofar n =
    do say $ "speak " ++ (replicate (sofar + (2^n)) 'a')
       (roomData, items) <- look
       let bit = case room_name roomData of
                   "Room With a Door" -> 1
                   "Junk Room" -> 0
       hackBit (sofar + bit*(2^n)) (n-1)

hackUpload :: Int -> IO ()
hackUpload i =
    do putStrLn "use uploader"
       firstHalf <- readFile "firsthalf"
       putStr firstHalf
       putStr $ show i
       secondHalf <- readFile "secondhalf"
       putStr secondHalf
       putStr "EOM"
       waitPrompt
       return ()

login :: IO ()
login =
    do wait "login: "
       putStrLn "howie"
       wait "password: "
       putStrLn "xyzzy"
       waitUmixPrompt
       return ()

start :: IO ()
start =
    do putStrLn "adventure"
       waitPrompt
       return ()

say :: String -> IO String
say s =
    do putStrLn s
       waitPrompt

waitUmixPrompt :: IO String
waitUmixPrompt =
    do wait "% "

waitPrompt :: IO String
waitPrompt =
    do wait ">: "

wait :: String -> IO String
wait what =
    wait' what ""

wait' :: String -> String -> IO String
wait' what s =
    let lenS = length s
        lenWhat = length what
    in if (lenS >= lenWhat) && (drop (lenS-lenWhat) s == what)
       then return s
       else do ch <- getChar
               wait' what (s ++ [ch])

prelude :: IO ()
prelude = play "prelude.txt"

play :: String -> IO ()
play file =
    do contents <- readFile ("../../doc/adventure/" ++ file)
       sequence_ $ map say (lines contents)

look :: IO (RoomData, [Item])
look =
    do txt <- say "look"
       case readsRoom txt of
         [(r, "")] -> return r
         _ -> error "readsRoom failed"

go :: Dir -> IO ()
go dir =
    do say ("go " ++ showDir dir)
       return ()

lookItems :: [Item] -> IO [ItemData]
lookItems items =
    sequence $ map lookItem items

lookItem :: Item -> IO ItemData
lookItem item =
    do txt <- say $ "look " ++ item
       case readsItem item txt of
         [(i, "")] -> return i
         _ -> error "readsItem failed"

solve :: Noun -> IO ()
solve noun =
    do (roomData, items) <- look
       itemDatas <- lookItems items
       let commands = solveCommands (Whole noun) itemDatas
       sequence_ $ map say commands

goAndSolve :: Noun -> [Dir] -> IO ()
goAndSolve noun [] = solve noun
goAndSolve noun (d:ds) =
    do go d
       goAndSolve noun ds
       go (opposite d)
