#include "State.hpp"
#include <stdio.h>
#include <iostream>

using namespace std;

void exec(State &state)
{
  while(true) {
    Platter pc = state.m_pc;
    state.m_pc++;

    vector<Platter> &imem = *(state.m_array0);
    Platter inst = imem[pc];

    Platter op = (inst & 0xf0000000) >> 28;

    //printf("pc = %d, inst = %08x, op = %d\n", pc, inst, op);

    if (op == 13) {
      Platter ra = (inst & 0x0e000000) >> 25;
      Platter val = (inst & 0x01ffffff);
      state.m_regs[ra] = val;
    } else {
      Platter ra = (inst & 0x000001c0) >> 6;
      Platter rb = (inst & 0x00000038) >> 3;
      Platter rc = (inst & 0x00000007);

      switch (op) {
      case 0: // Conditional Move
        {
          if (state.m_regs[rc] != 0) {
            state.m_regs[ra] = state.m_regs[rb];
          }
        }
        break;
      case 1: // Array Index
        {
          Platter arrayId = state.m_regs[rb];
          vector<Platter> &mem = arrayId ? (*((vector<Platter> *)arrayId)) : *(state.m_array0);
          state.m_regs[ra] = mem[state.m_regs[rc]];
        }
        break;
      case 2: // Array Amendment
        {
          Platter arrayId = state.m_regs[ra];
          vector<Platter> &mem = arrayId ? (*((vector<Platter> *)arrayId)) : *(state.m_array0);
          mem[state.m_regs[rb]] = state.m_regs[rc];
        }
        break;
      case 3: // Addition
        {
          state.m_regs[ra] = state.m_regs[rb] + state.m_regs[rc];
        }
        break;
      case 4: // Multiplication
        {
          state.m_regs[ra] = state.m_regs[rb] * state.m_regs[rc];
        }
        break;
      case 5: // Division
        {
          if (state.m_regs[rc] == 0) {
            printf("***Division by zero***\n");
            exit(1);
          }
          state.m_regs[ra] = state.m_regs[rb] / state.m_regs[rc];
        }
        break;
      case 6: // Non-And
        {
          state.m_regs[ra] = ~(state.m_regs[rb] & state.m_regs[rc]);
        }
        break;
      case 7: // Halt
        {
          printf("***Halted***\n");
          exit(0);
        }
        break;
      case 8: // Allocation
        {
          vector<Platter> *pVec = new vector<Platter>;
          pVec->resize(state.m_regs[rc]);
          Platter arrayId = (Platter)(pVec);
          state.m_regs[rb] = arrayId;
        }
        break;
      case 9: // Abandonment
        {
          Platter arrayId = state.m_regs[rc];
          delete ((vector<Platter> *)arrayId);
        }
        break;
      case 10: // Output
        {
          char ch = state.m_regs[rc] & 0xff;
          cout << ch;
	  fflush(0);
        }
        break;
      case 11: // Input
        {
          // WARNING: No EOF handling!
          //char ch = '\0';
          //cin >> ch;
	  char ch;
	  bool char_set = false;
	  while (!char_set && !state.m_inputs.empty()) {
            ifstream *is = state.m_inputs.front();
	    ch = is->get();
	    if (is->fail()) {
              state.m_inputs.pop_front();
	      delete is;
	    } else {
	      char_set = true;
	    }
	  }
          if (!char_set) ch = getchar();
          state.m_regs[rc] = ch;
        }
        break;
      case 12: // Load Program
        {
          Platter nArr = state.m_regs[rb];
          if (nArr != 0) {
            delete state.m_array0;
            state.m_array0 = new vector<Platter>(*((vector<Platter> *)nArr));
          }
          state.m_pc = state.m_regs[rc];
        }
        break;
      }
    }
  }
}
