module Main where

import System.IO
import System.Exit
import System.Environment
import Data.Word
import Data.Char
import qualified Data.Map as Map
import Data.Array.MArray
import Data.Bits
import Control.Monad
import UniversalMachine
import Exec
import Flags

debug :: Bool
debug = True

main :: IO ()
main =
    do args <- getArgs
       (flags, files) <- parse_flags args
       case files of
         [f] -> main' flags f
         _ -> usage

main' :: Flags -> String -> IO ()
main' flags progFile =
    do progChars <- readFile progFile
       let progPlatters = charsToPlatters progChars
           len = length progPlatters
       m0 <- new_machine
       progArray <- newListArray
                    (fromIntegral 0, fromIntegral (len - 1))
                    progPlatters
       let m1 = m0 { arrays = Map.insert 0 progArray (arrays m0) }
       hSetBuffering stdout NoBuffering
       hSetBuffering stderr LineBuffering
       exec flags m1
       return ()

charsToPlatters :: [Char] -> [Platter]
charsToPlatters [] = []
charsToPlatters (a:b:c:d:rest) =
    (charsToPlatter a b c d):(charsToPlatters rest)
charsToPlatters _ = error "Input size is not a multiple of 4 bytes"

charsToPlatter :: Char -> Char -> Char -> Char -> Platter
charsToPlatter a b c d =
    let a' = charToPlatter a
        b' = charToPlatter b
        c' = charToPlatter c
        d' = charToPlatter d
    in (a' `shiftL` 24) .|.
       (b' `shiftL` 16) .|.
       (c' `shiftL` 8) .|.
       d'

charToPlatter :: Char -> Platter
charToPlatter = fromIntegral . ord

usage :: IO ()
usage =
    do hPutStrLn stderr "Usage: interp <program>"
       exitWith (ExitFailure 1)
