#include <stdlib.h>
#include <stdio.h>
#include "jit.h"
#include "codegen.h"

void readreg(FILE *f, const char *indent, char r)
{
  fprintf(f, "%sswitch (%c) {\n", indent, r);
  fprintf(f, "%scase 0: r%c = r0; break;\n", indent, r);
  fprintf(f, "%scase 1: r%c = r1; break;\n", indent, r);
  fprintf(f, "%scase 2: r%c = r2; break;\n", indent, r);
  fprintf(f, "%scase 3: r%c = r3; break;\n", indent, r);
  fprintf(f, "%scase 4: r%c = r4; break;\n", indent, r);
  fprintf(f, "%scase 5: r%c = r5; break;\n", indent, r);
  fprintf(f, "%scase 6: r%c = r6; break;\n", indent, r);
  fprintf(f, "%scase 7: r%c = r7; break;\n", indent, r);
  fprintf(f, "%sdefault: r%c = 0;\n", indent, r);
  fprintf(f, "%s}\n", indent);
}

void savereg(FILE *f, const char *indent, char r)
{
  fprintf(f, "%sswitch (%c) {\n", indent, r);
  fprintf(f, "%scase 0: r0 = r%c; break;\n", indent, r);
  fprintf(f, "%scase 1: r1 = r%c; break;\n", indent, r);
  fprintf(f, "%scase 2: r2 = r%c; break;\n", indent, r);
  fprintf(f, "%scase 3: r3 = r%c; break;\n", indent, r);
  fprintf(f, "%scase 4: r4 = r%c; break;\n", indent, r);
  fprintf(f, "%scase 5: r5 = r%c; break;\n", indent, r);
  fprintf(f, "%scase 6: r6 = r%c; break;\n", indent, r);
  fprintf(f, "%scase 7: r7 = r%c; break;\n", indent, r);
  fprintf(f, "%s}\n", indent);
}

void codegen(unsigned int *imem,
             unsigned int imemsz,
             const unsigned int *instrs,
             unsigned int numinstrs,
             const unsigned int *targets)
{
  FILE *f = fopen("exec.c", "w");
  if (!f) {
    fprintf(stderr, "Could not open exec.c for write\n");
    exit(1);
  }

  fprintf(f, "#include <stdlib.h>\n");
  fprintf(f, "#include <stdio.h>\n");
  fprintf(f, "#include <string.h>\n");
  fprintf(f, "\n");
  fprintf(f, "typedef unsigned int *(*mo_alloc_t)(unsigned int sz);\n");
  fprintf(f, "typedef void (*mo_free_t)(unsigned int *p);\n");
  fprintf(f, "typedef unsigned int (*mo_size_t)(unsigned int *p);\n");
  fprintf(f, "\n");
  fprintf(f, "typedef struct {\n");
  fprintf(f, "  mo_alloc_t alloc;\n");
  fprintf(f, "  mo_free_t free;\n");
  fprintf(f, "  mo_size_t size;\n");
  fprintf(f, "} memops;\n");
  fprintf(f, "\n");
  fprintf(f, "int exec(unsigned int **pimem,\n");
  fprintf(f, "         unsigned int *pimemsz,\n");
  fprintf(f, "         unsigned int *ppc,\n");
  fprintf(f, "         unsigned int *pregs,\n");
  fprintf(f, "         memops *mo)\n");
  fprintf(f, "{\n");
  fprintf(f, "  unsigned int interpretcount = %u;\n", REPROFILECOUNT);
  fprintf(f, "  int forcerecompile = 0;\n");
  fprintf(f, "\n");
  fprintf(f, "  unsigned int inst = 0;\n");
  fprintf(f, "  unsigned int op = 0;\n");
  fprintf(f, "\n");
  fprintf(f, "  unsigned int *imem = *pimem;\n");
  fprintf(f, "  unsigned int imemsz = *pimemsz;\n");
  fprintf(f, "  unsigned int pc = *ppc;\n");

  unsigned int i;
  for (i = 0; i < 8; i++) {
    fprintf(f, "  unsigned int r%u = pregs[%u];\n", i, i);
  }

  fprintf(f, "\n");
  fprintf(f, "  int *jumptbl = malloc(imemsz * sizeof(int));\n");
  fprintf(f, "  unsigned int i;\n");
  fprintf(f, "  for (i = 0; i < imemsz; i++) {\n");
  fprintf(f, "    jumptbl[i] = &&interpret - &&start;\n");
  fprintf(f, "  }\n");
  fprintf(f, "\n");

  for (i = 0; i < numinstrs; i++) {
    unsigned int pc = instrs[i];
    if (targets[pc]) {
      fprintf(f, "  jumptbl[%u] = &&label%u - &&start;\n", pc, pc);
    }
  }

  fprintf(f, "\n");
  fprintf(f, "  goto *(&&start + jumptbl[pc]);\n");
  fprintf(f, "\n");

  fprintf(f, " start:\n");

  for (i = 0; i < numinstrs; i++) {
    unsigned int pc = instrs[i];
    if (targets[pc]) {
      fprintf(f, " label%u:\n", pc);
    }

    unsigned int inst = imem[pc];
    unsigned int op = (inst & 0xf0000000) >> 28;

    if (op == 13) { /* Orthography */
      unsigned int a = (inst & 0x0e000000) >> 25;
      unsigned int val = (inst & 0x01ffffff);
      fprintf(f, "  r%u = 0x%08x;\n", a, val);
    } else {
      unsigned int a = (inst & 0x000001c0) >> 6;
      unsigned int b = (inst & 0x00000038) >> 3;
      unsigned int c = (inst & 0x00000007);

      switch (op) {
      case 0: /* Conditional Move */
        fprintf(f, "  if (r%u) r%u = r%u;\n", c, a, b);
        break;
      case 1: /* Array Index */
        fprintf(f, "  r%u = (r%u ? (unsigned int *)(r%u) : imem)[r%u];\n", a, b, b, c);
        break;
      case 2: /* Array Amendment */
#if DETECT_SELF_REWRITING_CODE
        fprintf(f, "  if (r%u) {\n", a);
        fprintf(f, "    ((unsigned int *)(r%u))[r%u] = r%u;\n", a, b, c);
        fprintf(f, "  } else {\n");
        fprintf(f, "    imem[r%u] = r%u;\n", b, c);
        fprintf(f, "    if (jumptbl[r%u] != &&interpret - &&start) {\n", b);
        fprintf(f, "      pc = %u;\n", pc+1);
        fprintf(f, "      goto self_rewriting_code;\n");
        fprintf(f, "    }\n");
        fprintf(f, "  }\n");
#else
        fprintf(f, "  (r%u ? (unsigned int *)(r%u) : imem)[r%u] = r%u;\n", a, a, b, c);
#endif
        break;
      case 3: /* Addition */
        fprintf(f, "  r%u = r%u + r%u;\n", a, b, c);
        break;
      case 4: /* Multiplication */
        fprintf(f, "  r%u = r%u * r%u;\n", a, b, c);
        break;
      case 5: /* Division */
        fprintf(f, "  r%u = r%u / r%u;\n", a, b, c);
        break;
      case 6: /* Non-And */
        fprintf(f, "  r%u = ~(r%u & r%u);\n", a, b, c);
        break;
      case 7: /* Halt */
        fprintf(f, "  exit(0);\n");
        break;
      case 8: /* Allocation */
        fprintf(f, "  r%u = (unsigned int)mo->alloc(r%u);\n", b, c);
        break;
      case 9: /* Abandonment */
        fprintf(f, "  mo->free((unsigned int *)r%u);\n", c);
        break;
      case 10: /* Output */
        fprintf(f, "  putchar(r%u);\n", c);
        break;
      case 11: /* Input */
        fprintf(f, "  r%u = getchar();\n", c);
        break;
      case 12: /* Load Program */
        fprintf(f, "  if (r%u) {\n", b);
        fprintf(f, "    pc = %u;\n", pc);
        fprintf(f, "    goto interpret;\n");
        fprintf(f, "  } else {\n");
        fprintf(f, "    pc = r%u;\n", c);
        fprintf(f, "    goto *(&&start + jumptbl[pc]);\n");
        fprintf(f, "  }\n");
        break;
      default:
        fprintf(f, "  fprintf(stderr, \"*** Bad opcode ***\\n\");\n");
        fprintf(f, "  exit(1);\n");
      }
    }

    if (i == numinstrs-1 ||
        instrs[i+1] != pc+1) {
      fprintf(f, "  pc = %u;\n", pc+1);
      fprintf(f, "  goto *(&&start + jumptbl[pc]);\n");
    }
  }

  fprintf(f, "\n");
  fprintf(f, " interpret:\n");
  fprintf(f, "  if (interpretcount == 0) {\n");
  fprintf(f, "    goto recompile;\n");
  fprintf(f, "  }\n");
  fprintf(f, "  interpretcount--;\n");
  fprintf(f, "\n");
  fprintf(f, "  inst = imem[pc++];\n");
  fprintf(f, "  op = (inst & 0xf0000000) >> 28;\n");
  fprintf(f, "\n");
  fprintf(f, "  if (op == 13) { /* Orthography */\n");
  fprintf(f, "    unsigned int a = (inst & 0x0e000000) >> 25;\n");
  fprintf(f, "    unsigned int val = (inst & 0x01ffffff);\n");
  fprintf(f, "    unsigned int ra;\n");
  fprintf(f, "    ra = val;\n");
  savereg(f, "    ", 'a');
  fprintf(f, "  } else {\n");
  fprintf(f, "    unsigned int a = (inst & 0x000001c0) >> 6;\n");
  fprintf(f, "    unsigned int b = (inst & 0x00000038) >> 3;\n");
  fprintf(f, "    unsigned int c = (inst & 0x00000007);\n");
  fprintf(f, "    unsigned int ra, rb, rc;\n");
  fprintf(f, "\n");
  fprintf(f, "    switch (op) {\n");
  fprintf(f, "    case 0: /* Conditional Move */\n");
  readreg(f, "      ", 'c');
  readreg(f, "      ", 'b');
  fprintf(f, "      if (rc != 0) {\n");
  fprintf(f, "        ra = rb;\n");
  savereg(f, "        ", 'a');
  fprintf(f, "      }\n");
  fprintf(f, "      break;\n");
  fprintf(f, "    case 1: /* Array Index */\n");
  readreg(f, "      ", 'b');
  readreg(f, "      ", 'c');
  fprintf(f, "      ra = (rb ? (unsigned int *)(rb) : imem)[rc];\n");
  savereg(f, "      ", 'a');
  fprintf(f, "      break;\n");
  fprintf(f, "    case 2: /* Array Amendment */\n");
  readreg(f, "      ", 'a');
  readreg(f, "      ", 'b');
  readreg(f, "      ", 'c');
#if DETECT_SELF_REWRITING_CODE
  fprintf(f, "      if (ra) {\n");
  fprintf(f, "        ((unsigned int *)(ra))[rb] = rc;\n");
  fprintf(f, "      } else {\n");
  fprintf(f, "        imem[rb] = rc;\n");
  fprintf(f, "        if (jumptbl[rb] != &&interpret - &&start) {\n");
  fprintf(f, "          goto self_rewriting_code;\n");
  fprintf(f, "        }\n");
  fprintf(f, "      }\n");
#else
  fprintf(f, "      (ra ? (unsigned int *)(ra) : imem)[rb] = rc;\n");
#endif
  fprintf(f, "      break;\n");
  fprintf(f, "    case 3: /* Addition */\n");
  readreg(f, "      ", 'b');
  readreg(f, "      ", 'c');
  fprintf(f, "      ra = rb + rc;\n");
  savereg(f, "      ", 'a');
  fprintf(f, "      break;\n");
  fprintf(f, "    case 4: /* Multiplication */\n");
  readreg(f, "      ", 'b');
  readreg(f, "      ", 'c');
  fprintf(f, "      ra = rb * rc;\n");
  savereg(f, "      ", 'a');
  fprintf(f, "      break;\n");
  fprintf(f, "    case 5: /* Division */\n");
  readreg(f, "      ", 'b');
  readreg(f, "      ", 'c');
  fprintf(f, "      ra = rb / rc;\n");
  savereg(f, "      ", 'a');
  fprintf(f, "      break;\n");
  fprintf(f, "    case 6: /* Non-And */\n");
  readreg(f, "      ", 'b');
  readreg(f, "      ", 'c');
  fprintf(f, "      ra = ~(rb & rc);\n");
  savereg(f, "      ", 'a');
  fprintf(f, "      break;\n");
  fprintf(f, "    case 7: /* Halt */\n");
  fprintf(f, "      exit(0);\n");
  fprintf(f, "      break;\n");
  fprintf(f, "    case 8: /* Allocation */\n");
  readreg(f, "      ", 'c');
  fprintf(f, "      rb = (unsigned int)mo->alloc(rc);\n");
  savereg(f, "      ", 'b');
  fprintf(f, "      break;\n");
  fprintf(f, "    case 9: /* Abandonment */\n");
  readreg(f, "      ", 'c');
  fprintf(f, "      mo->free((unsigned int *)rc);\n");
  fprintf(f, "      break;\n");
  fprintf(f, "    case 10: /* Output */\n");
  readreg(f, "      ", 'c');
  fprintf(f, "      putchar(rc);\n");
  fprintf(f, "      break;\n");
  fprintf(f, "    case 11: /* Input */\n");
  fprintf(f, "      rc = getchar();\n");
  savereg(f, "      ", 'c');
  fprintf(f, "      break;\n");
  fprintf(f, "    case 12: /* Load Program */\n");
  readreg(f, "      ", 'b');
  readreg(f, "      ", 'c');
  fprintf(f, "      if (rb) {\n");
  fprintf(f, "        free(imem);\n");
  fprintf(f, "        imemsz = *(((unsigned int *)(rb)) - 1);\n");
  fprintf(f, "        imem = malloc(imemsz * sizeof(unsigned int));\n");
  fprintf(f, "        memcpy(imem, (unsigned int *)(rb), imemsz * sizeof(unsigned int));\n");
  fprintf(f, "        pc = rc;\n");
  fprintf(f, "        goto force_recompile;\n");
  fprintf(f, "      } else {\n");
  fprintf(f, "        pc = rc;\n");
  fprintf(f, "      }\n");
  fprintf(f, "      break;\n");
  fprintf(f, "    default:\n");
  fprintf(f, "      fprintf(stderr, \"*** Bad opcode ***\\n\");\n");
  fprintf(f, "      exit(1);\n");
  fprintf(f, "    }\n");
  fprintf(f, "  }\n");
  fprintf(f, "  goto *(&&start + jumptbl[pc]);\n");
  fprintf(f, "\n");
#if DETECT_SELF_REWRITING_CODE
  fprintf(f, " self_rewriting_code:\n");
  fprintf(f, "  printf(\"*** SELF REWRITING CODE! RECOMPILING ***\\n\");\n");
#endif
  fprintf(f, "\n");
  fprintf(f, " force_recompile:\n");
  fprintf(f, "  forcerecompile = 1;\n");
  fprintf(f, "\n");
  fprintf(f, " recompile:\n");
  fprintf(f, "  free(jumptbl);\n");
  fprintf(f, "  *pimem = imem;\n");
  fprintf(f, "  *pimemsz = imemsz;\n");
  fprintf(f, "  *ppc = pc;\n");

  for (i = 0; i < 8; i++) {
    fprintf(f, "  pregs[%u] = r%u;\n", i, i);
  }

  fprintf(f, "  return forcerecompile;\n");
  fprintf(f, "}\n");

  fclose(f);
}
