#include <stdlib.h>
#include <string.h>
#include "fastalloc.h"

unsigned int *buckets[16];

void fa_init()
{
  unsigned int i;
  for (i = 0; i < 16; i++) {
    buckets[i] = 0;
  }
}

int getbucket(unsigned int sz)
{
  unsigned int i;
  for (i = 0; i < 16; i++) {
    if (sz <= (1 << i)) {
      return i;
    }
  }
  return -1;
}

void fillbucket(int bucket)
{
  unsigned int *p = malloc((1 << 16) * sizeof(unsigned int));
  unsigned int sz = 1 << bucket;
  unsigned int i;
  unsigned int *last = 0;
  for (i = 0; i < (1<<16); i+=sz) {
    *(unsigned int **)(p+i) = last;
    last = p+i;
  }
  buckets[bucket] = last;
}

unsigned int *fa_alloc(unsigned int sz)
{
  unsigned int realsz = sz + 1;
  int bucket = getbucket(realsz);
  if (bucket == -1) {
    unsigned int *p = calloc(realsz, sizeof(unsigned int));
    *p = sz;
    return p+1;
  } else {
    if (!buckets[bucket]) {
      fillbucket(bucket);
    }
    unsigned int *p = buckets[bucket];
    buckets[bucket] = *(unsigned int **)p;
    memset(p+1, 0, sz * sizeof(unsigned int));
    *p = sz;
    return p+1;
  }
}

void fa_free(unsigned int *p)
{
  unsigned int sz = *(p-1);
  unsigned int realsz = sz + 1;
  int bucket = getbucket(realsz);
  if (bucket == -1) {
    free(p-1);
  } else {
    *(unsigned int **)(p-1) = buckets[bucket];
    buckets[bucket] = p-1;
  }
}

unsigned int fa_size(unsigned int *p)
{
  return *(p-1);
}
