#include <stdlib.h>
#include <stdio.h>
#include "file.h"

void readfile(const char *filename,
              unsigned int **pimem,
              unsigned int *pimemsz)
{
  FILE *infile = fopen(filename, "rb");
  if (!infile) {
    fprintf(stderr, "Could not open program file %s\n", filename);
    exit(1);
  }

  size_t sz = 0;

  unsigned int *imem = 0;
  while (!feof(infile)) {
    static const int BUFSIZE = 65536;
    unsigned char buf[BUFSIZE];

    size_t bytecount = fread(buf, sizeof(unsigned char), BUFSIZE, infile);

    if (bytecount == 0) {
      fclose(infile);
      fprintf(stderr, "Error reading from program file\n");
      exit(1);
    }

    if (bytecount % 4 != 0) {
      fclose(infile);
      fprintf(stderr, "Program file size is not a multiple of 4\n");
      exit(1);
    }

    size_t count = bytecount / 4;

    imem = realloc(imem, (sz + count) * sizeof(unsigned int));

    if (!imem) {
      fclose(infile);
      fprintf(stderr, "Out of memory reading program file\n");
      exit(1);
    }

    size_t i;
    for (i = 0; i < count; i++) {
      imem[sz + i] =
        (((unsigned int)(buf[i * 4 + 0])) << 24) |
        (((unsigned int)(buf[i * 4 + 1])) << 16) |
        (((unsigned int)(buf[i * 4 + 2])) << 8) |
        (((unsigned int)(buf[i * 4 + 3])));
    }

    sz += count;
  }

  fclose(infile);

  if (!imem) {
    fprintf(stderr, "Program file is empty\n");
    exit(1);
  }

  *pimem = imem;
  *pimemsz = sz;
}
