#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "jit.h"
#include "profile.h"
#include "fastalloc.h"

double profile(int debug,
               unsigned int **pimem,
               unsigned int *pimemsz,
               unsigned int *ppc,
               unsigned int *pregs,
               unsigned int **pinstrs,
               unsigned int *pnuminstrs,
               const unsigned int *oldinstrs,
               unsigned int oldnuminstrs,
               unsigned int **pcounts,
               unsigned int **ptargets)
{
  unsigned int inst = 0;
  unsigned int op = 0;

  unsigned int *imem = *pimem;
  unsigned int imemsz = *pimemsz;
  unsigned int pc = *ppc;
  unsigned int r[8];
  unsigned int i;
  for (i = 0; i < 8; i++) {
    r[i] = pregs[i];
  }

  unsigned int *counts = *pcounts;
  if (counts) {
    for (i = 0; i < imemsz; i++) {
      counts[i] = (counts[i] / 10) * 9;
    }
  }
  if (!counts) {
    counts = calloc(imemsz, sizeof(unsigned int));
  }

  unsigned int *targets = *ptargets;
  if (!targets) {
    targets = calloc(imemsz, sizeof(unsigned int));
  }

  unsigned int count = 0;
  while(count < PROFILECOUNT) {
    counts[pc]++;

    inst = imem[pc++];
    op = (inst & 0xf0000000) >> 28;

    if (op == 13) { /* Orthography */
      unsigned int ra = (inst & 0x0e000000) >> 25;
      unsigned int val = (inst & 0x01ffffff);
      r[ra] = val;
    } else {
      unsigned int ra = (inst & 0x000001c0) >> 6;
      unsigned int rb = (inst & 0x00000038) >> 3;
      unsigned int rc = (inst & 0x00000007);

      switch (op) {
      case 0: /* Conditional Move */
        if (r[rc] != 0) r[ra] = r[rb];
        break;
      case 1: /* Array Index */
        r[ra] = (r[rb] ? (unsigned int *)(r[rb]) : imem)[r[rc]];
        break;
      case 2: /* Array Amendment */
        (r[ra] ? (unsigned int *)(r[ra]) : imem)[r[rb]] = r[rc];
        break;
      case 3: /* Addition */
        r[ra] = r[rb] + r[rc];
        break;
      case 4: /* Multiplication */
        r[ra] = r[rb] * r[rc];
        break;
      case 5: /* Division */
        r[ra] = r[rb] / r[rc];
        break;
      case 6: /* Non-And */
        r[ra] = ~(r[rb] & r[rc]);
        break;
      case 7: /* Halt */
        exit(0);
        break;
      case 8: /* Allocation */
        r[rb] = (unsigned int)fa_alloc(r[rc]);
        break;
      case 9: /* Abandonment */
        fa_free((unsigned int *)(r[rc]));
        break;
      case 10: /* Output */
        putchar(r[rc]);
        break;
      case 11: /* Input */
        r[rc] = getchar();
        break;
      case 12: /* Load Program */
        if (r[rb]) {
          free(imem);
          imemsz = *(((unsigned int *)(r[rb])) - 1);
          imem = malloc(imemsz * sizeof(unsigned int));
          memcpy(imem, (unsigned int *)(r[rb]), imemsz * sizeof(unsigned int));
          pc = r[rc];

          free(counts);
          counts = calloc(imemsz, sizeof(unsigned int));
          free(targets);
          targets = calloc(imemsz, sizeof(unsigned int));

          count = 0;
          continue;
        } else {
          pc = r[rc];
          targets[pc] = 1;
        }
        break;
      default:
        fprintf(stderr, "*** Bad opcode ***\n");
        exit(1);
      }
    }

    count++;
  }

  unsigned int alltotal = 0;
  for (i = 0; i < imemsz; i++) {
    alltotal += counts[i];
  }

  int numinstrs = 0;
  for (i = 0; i < imemsz; i++) {
    if ((double)counts[i] / (double)alltotal >= MINVALUE) {
      numinstrs++;
    }
  }

  unsigned int *instrs = 0;
  if (numinstrs > 0) {
    instrs = malloc(numinstrs * sizeof(unsigned int));
    unsigned int j = 0;
    for (i = 0; i < imemsz; i++) {
      if ((double)counts[i] / (double)alltotal >= MINVALUE) {
        instrs[j++] = i;
      }
    }
  }

  unsigned int total = 0;
  for (i = 0; i < numinstrs; i++) {
    total += counts[instrs[i]];
  }

  unsigned int oldtotal = 0;
  for (i = 0; i < oldnuminstrs; i++) {
    oldtotal += counts[oldinstrs[i]];
  }

  double value = 0.0;
  if (numinstrs > 0) {
    value = (((double)total / (double)alltotal) -
             ((double)oldtotal / (double)alltotal)) / (double)numinstrs;
  }

  if (debug) {
    printf("Expected hit ratio %.1f%%, from %u instructions, old ratio %.1f%%\n",
           100.0 * (double)total / (double)alltotal,
           numinstrs,
           100.0 * (double)oldtotal / (double)alltotal);
  }

  *pinstrs = instrs;
  *pnuminstrs = numinstrs;

  *pimem = imem;
  *pimemsz = imemsz;
  *ppc = pc;
  for (i = 0; i < 8; i++) {
    pregs[i] = r[i];
  }

  *pcounts = counts;
  *ptargets = targets;

  return value;
}
