module Main where

import Data.Char
import Data.Array
import Data.List

main :: IO ()
main =
    do input <- getContents
       case readsPlinko input of
         [(vals, "")] -> main' vals
         _ -> putStrLn "Parse failed"

main' :: [(Int, Int)] -> IO ()
main' valsList =
    do let len = length valsList
           vals = listArray (0, len-1) valsList
           moves = valsToMoves vals
           sorts = valsToSorts vals
           stage1in = listArray (0, len-1) (map fst valsList)
           stage1res = stage1 stage1in stage2in
           stage2in = listArray (0, len-1) (map snd (sort $ elems sorts))
           counts = stage2in
       --print $ stage1res
       --putStrLn ""
       --print stage2in
       putStr $ concatMap (showMoves len) stage1res
       --print $ stage1res
       --putStr $ unlines $ map show $ assocs sorts
       --print moves
       --putStr $ concatMap (showMove len) (greedy moves)

doTweaks :: Array Int Int -> Maybe ([Int], Int)
doTweaks counts =
    let theList = getTheList counts
    in doTweaks' counts theList

doTweaks' :: Array Int Int -> [[Int]] -> Maybe ([Int], Int)
doTweaks' _ [] = Nothing
doTweaks' counts (t:ts) =
    let counts' = applyTweaks counts t
    in case meet counts' of
         Just pos -> Just (t, pos)
         Nothing -> doTweaks' counts ts

getTheList :: Array Int Int -> [[Int]]
getTheList counts =
    let (low, high) = bounds counts
    in [[a, b] |
        a <- [low..high],
        b <- [low..high]]

applyTweaks :: Array Int Int -> [Int] -> Array Int Int
applyTweaks counts [] = counts
applyTweaks counts (i:is) =
    applyTweaks (tweak counts i) is

tweak :: Array Int Int -> Int -> Array Int Int
tweak counts n =
    let (low, high) = bounds counts
    in if (n >= low) && (n < high) &&
           (counts ! n > 1) && (counts ! (n+1) > 1)
       then adjustCounts counts [n]
       else counts

meet :: Array Int Int -> Maybe Int
meet counts =
    let (fromLeft, fromRight) = makeLeftRight counts
    in meet' fromLeft fromRight 0

meet' :: Array Int Int -> Array Int Int -> Int -> Maybe Int
meet' left right n =
    let (low, high) = bounds left
    in if n >= high
       then Nothing
       else if left ! n == right ! (n+1)
            then Just n
            else meet' left right (n+1)

makeLeftRight :: Array Int Int -> (Array Int Int, Array Int Int)
makeLeftRight counts =
    (left, right)
    where
      (low, high) = bounds counts
      left = listArray
             (low, high)
             ([counts ! 0] ++ map fnLeft [1..high])
      right = listArray
              (low, high)
              (map fnRight [0..high-1] ++ [counts ! high])
      fnLeft n =
          abs ((counts ! n) - (left ! (n-1)))
      fnRight n =
          abs ((counts ! n) - (right ! (n+1)))

stage1 :: Array Int Int -> Array Int Int -> [[Int]]
stage1 newOrder counts =
    let swaps = stage1' newOrder 0
        newOrder' = applySwaps newOrder swaps
    in if null swaps
       then stage2 counts
       else swaps:(stage1 newOrder' counts)

stage1' :: Array Int Int -> Int -> [Int]
stage1' newOrder start =
    let (low, high) = bounds newOrder
    in if start >= high
       then []
       else if (newOrder ! start) > (newOrder ! (start+1))
            then start:(stage1' newOrder (start+2))
            else stage1' newOrder (start+1)

applySwaps :: Array Int a -> [Int] -> Array Int a
applySwaps arr swaps =
    arr // (concatMap fn swaps)
    where
      fn n = [(n, arr ! (n+1)), (n+1, arr ! n)]

stage2 :: Array Int Int -> [[Int]]
stage2 counts =
    let swaps = stage2' counts 0
        counts' = adjustCounts counts swaps
    in if null swaps
       then case meet counts of
              Just pos -> stage3 counts pos
              Nothing -> error "Failed"
       else swaps:swaps:stage2 counts'

stage2' :: Array Int Int -> Int -> [Int]
stage2' counts start =
    let (low, high) = bounds counts
    in if start >= high
       then []
       else if ((counts ! start) > 1) && ((counts ! (start+1)) > 1)
            then start:(stage2' counts (start+2))
            else stage2' counts (start+1)

adjustCounts :: Array Int Int -> [Int] -> Array Int Int
adjustCounts arr swaps =
    arr // (concatMap fn swaps)
    where
      fn n = [(n, (arr ! n) - 1), (n+1, (arr ! (n+1)) - 1)]

stage3 :: Array Int Int -> Int -> [[Int]]
stage3 counts pos =
    stage3' counts pos 0

stage3' :: Array Int Int -> Int -> Int -> [[Int]]
stage3' counts pos start =
    let (low, high) = bounds counts
        swaps = [start]
        counts' = adjustCounts counts swaps
        swappedCounts = applySwaps counts' swaps
    in if start == pos
       then stage4 counts pos
       else if counts ! start == 0
            then stage3' counts pos (start+1)
            else if counts ! (start+1) == 1
                 then [swaps] ++ stage3' swappedCounts pos start ++ [swaps]
                 else swaps:swaps:stage3' counts' pos start

stage4 :: Array Int Int -> Int -> [[Int]]
stage4 counts pos =
    let (low, high) = bounds counts
    in stage4' counts pos high

stage4' :: Array Int Int -> Int -> Int -> [[Int]]
stage4' counts pos start =
    let (low, high) = bounds counts
        swaps = [start-1]
        counts' = adjustCounts counts swaps
        swappedCounts = applySwaps counts' swaps
    in if start == pos
       then []
       else if counts ! start == 0
            then stage4' counts pos (start-1)
            else if counts ! (start-1) == 1
                 then [swaps] ++ stage4' swappedCounts pos start ++ [swaps]
                 else swaps:swaps:stage4' counts' pos start

valsToSorts :: Array Int (Int, Int) -> Array Int (Int, Int)
valsToSorts vals =
    let (low, high) = bounds vals
    in listArray (bounds vals) (map (valToSort vals) [0..high])

valToSort :: Array Int (Int, Int) -> Int -> (Int, Int)
valToSort vals n =
    let (y, z) = vals ! n
        valsList = map fst $ elems vals
        (left, rightPlusN) = span (/= y) valsList
        right = tail rightPlusN
        leftCount = length $ filter (> y) left
        rightCount = length $ filter (< y) right
    in (y, z - rightCount)


showMove :: Int -> Int -> String
showMove len n =
    map fn [0..len-1] ++ "\n"
    where
      fn x | x == n = '>'
      fn x | x == n+1 = '<'
      fn _ = '|'

showMoves :: Int -> [Int] -> String
showMoves len swaps =
    map fn [0..len-1] ++ "\n"
    where
      fn n | (n `elem` swaps) = '>'
      fn n | ((n-1) `elem` swaps) = '<'
      fn _ = '|'

greedy :: Array Int (Int, Int) -> [Int]
greedy moves =
    let (low, high) = bounds moves
        scores = map (getScore moves) [low..(high-1)]
        (bestScore, best) = maximum (zip scores [low..(high-1)])
        moves' = applyMove best moves
    in if bestScore == 0
       then []
       else best:(greedy moves')

applyMove :: Int -> Array Int (Int, Int) -> Array Int (Int, Int)
applyMove n moves =
    let (left0, right0) = moves ! n
        (left1, right1) = moves ! (n+1)
    in if right0 <= 0 || left1 <= 0
       then error $ "Failed: " ++ show moves
       else
    moves // [(n, (left1 - 1, right1)), (n+1, (left0, right0 - 1))]

getScore :: Array Int (Int, Int) -> Int -> Int
getScore moves n =
    (snd (moves ! n)) + (fst (moves ! (n+1)))

valsToMoves :: Array Int (Int, Int) -> Array Int (Int, Int)
valsToMoves vals =
    let (low, high) = bounds vals
    in listArray (bounds vals) (map (valToMove vals) [0..high])

valToMove :: Array Int (Int, Int) -> Int -> (Int, Int)
valToMove vals n =
    let (y, z) = vals ! n
    in (n+z - y, z)

readsPlinko :: ReadS[(Int, Int)]
readsPlinko = readsPlinko' 0

readsPlinko' :: Int -> ReadS [(Int, Int)]
readsPlinko' _ "" = [([], "")]
readsPlinko' n s =
    [(((y,z):rest),"") |
     (n, s1) <- readsInt s,
     (' ', s2) <- readsChar s1,
     ('-', s3) <- readsChar s2,
     ('>', s4) <- readsChar s3,
     (' ', s5) <- readsChar s4,
     ('(', s6) <- readsChar s5,
     (y, s7) <- readsInt s6,
     (',', s8) <- readsChar s7,
     (z, s9) <- readsInt s8,
     (')', s10) <- readsChar s9,
     ('\n', s') <- readsChar s10,
     (rest, "") <- readsPlinko' (n+1) s']

readsChar :: ReadS Char
readsChar [] = []
readsChar (c:cs) = [(c, cs)]

readsDigit :: ReadS Int
readsDigit [] = []
readsDigit (d:s') | isDigit d = [(digitToInt d, s')]
                  | otherwise = []

readsInt :: ReadS Int
readsInt s =
    [(i, s') |
     (d, s1) <- readsDigit s,
     (i, s') <- readsIntAux d s1]

readsIntAux :: Int -> ReadS Int
readsIntAux i "" = [(i, "")]
readsIntAux i s@(d:s') | isDigit d = readsIntAux (i * 10 + digitToInt d) s'
                       | otherwise = [(i, s)]
